#include <stdlib.h>
#include "png.h"

//BITMAPFILEHEADER  ptr index
#define BF_TYPE 0
#define BF_SIZE 2
#define BF_RESERVED1 6
#define BF_RESERVED2 8
#define BF_OFFBITS 10

//WINBMPINFOHEADER  ptr index
#define BI_SIZE 14
#define BI_WIDTH 18
#define BI_HEIGHT 22
#define BI_PLANES 26
#define BI_BITCOUNT 28
#define BI_COMPRESSION 30
#define BI_SIZEIMAGE 34
#define BI_XPELSPERMETER 38
#define BI_YPELSPERMETER 42
#define BI_CLRUSED 46
#define BI_CLRIMPORTANT 50

//Palette entry  ptr index
#define BF_PALETTE 54

//DIB  ptr index
#define BF_DIBP8 1078
#define BF_DIBP24 54

#undef TRUE
#undef FALSE
#define TRUE 1
#define FALSE 0

//low 0 - 9 high
static int complevel = 9;

static int interlace = 0;
static int filters   = 0;
static int bpHYs = FALSE;

//png palette
static png_color pngpal[256];
static png_text pngtxt;

static unsigned char CheckBitmapfileHeader(unsigned char *pBitmapImage)
{
	//Check bitmap file header.
	if(!pBitmapImage ||
		*(unsigned short *)(&pBitmapImage[BF_TYPE]) != 0x4D42 ||
		*(unsigned short *)(&pBitmapImage[BF_RESERVED1]) != 0 ||
		*(unsigned short *)(&pBitmapImage[BF_RESERVED2]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_SIZE]) != 40 ||
		*(unsigned short *)(&pBitmapImage[BI_PLANES]) != 1 ||
		*(unsigned long *)(&pBitmapImage[BI_COMPRESSION]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_XPELSPERMETER]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_YPELSPERMETER]) != 0 ||
		!((*(unsigned long *)(&pBitmapImage[BF_OFFBITS]) == 54 && *(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 24 &&
		*(unsigned long *)(&pBitmapImage[BI_CLRUSED]) == 0 && *(unsigned long *)(&pBitmapImage[BI_CLRIMPORTANT]) == 0) ||
		(*(unsigned long *)(&pBitmapImage[BF_OFFBITS]) == 1078 && *(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 8 &&
		*(unsigned long *)(&pBitmapImage[BI_CLRUSED]) == 256 && *(unsigned long *)(&pBitmapImage[BI_CLRIMPORTANT]) == 256)))
		return FALSE;
	return TRUE;
}

unsigned char bmp2png(unsigned char *pBitmapImage, FILE *fp, char *png_comment_key_str, char *png_comment_text_str)
{
	if(!CheckBitmapfileHeader(pBitmapImage) || !fp)
		return FALSE;

	png_structp ppng = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if(!ppng)
		return FALSE;
	png_infop pipng = png_create_info_struct(ppng);
	if(!pipng) {
		png_destroy_write_struct(&ppng, NULL);
		return FALSE;
	}
	if(setjmp(ppng->jmpbuf)){
		png_destroy_write_struct(&ppng, &pipng);
		return FALSE;
	}
	png_init_io(ppng, fp);
	png_set_compression_level(ppng, complevel);
	if( filters!=0 )
		png_set_filter(ppng, PNG_FILTER_TYPE_BASE, filters);
	
	int bit_depth;
	int color_type;

	if(*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 24){
		bit_depth  = 8;
		color_type = PNG_COLOR_TYPE_RGB;
		png_set_compression_mem_level(ppng, MAX_MEM_LEVEL);
	}
	else{
		bit_depth  = (int)(*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]));
		color_type = PNG_COLOR_TYPE_PALETTE;
		unsigned long ClrUsed = *(unsigned long *)(&pBitmapImage[BI_CLRUSED]);
		for(unsigned long i = 0; i < ClrUsed; i++)
		{
			pngpal[i].red = pBitmapImage[BF_PALETTE + i * 4 + 2];
			pngpal[i].green = pBitmapImage[BF_PALETTE + i * 4 + 1];
			pngpal[i].blue = pBitmapImage[BF_PALETTE + i * 4];
		}
		png_set_PLTE(ppng, pipng, pngpal, (int)ClrUsed);
	}
	png_set_IHDR(ppng, pipng, *(unsigned long *)(&pBitmapImage[BI_WIDTH]), *(unsigned long *)(&pBitmapImage[BI_HEIGHT]), bit_depth,
	              color_type, interlace ? PNG_INTERLACE_ADAM7 : PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
	              PNG_FILTER_TYPE_DEFAULT);

	if(bpHYs && (*(unsigned long *)(&pBitmapImage[BI_XPELSPERMETER])) && (*(unsigned long *)(&pBitmapImage[BI_YPELSPERMETER])))
	{
		png_set_pHYs(ppng, pipng, (png_int_32)(*(unsigned long *)(&pBitmapImage[BI_XPELSPERMETER])),
			(png_int_32)(*(unsigned long *)(&pBitmapImage[BI_YPELSPERMETER])),
			PNG_RESOLUTION_METER);
	}

	//png comment.
	pngtxt.compression = PNG_TEXT_COMPRESSION_NONE;
	pngtxt.key = png_comment_key_str;
	pngtxt.text = png_comment_text_str;
	png_set_text(ppng, pipng, &pngtxt, 1);

	png_write_info(ppng, pipng);
	if(*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 24)
		png_set_bgr(ppng);

	unsigned long dibheight = *(unsigned long *)(&pBitmapImage[BI_HEIGHT]);
	unsigned char *pdibdata = pBitmapImage + (*(unsigned long *)(&pBitmapImage[BF_OFFBITS]));

	//set ptr of DIB line.
	png_bytepp pimgline = (png_bytep *) malloc(sizeof(png_bytep) * dibheight);

	if(!pimgline)
		return FALSE;
	
	unsigned long dibPitch = ((*(unsigned long *)(&pBitmapImage[BI_SIZEIMAGE])) / dibheight);

	for(unsigned long diblinenum = 0; diblinenum < dibheight; diblinenum++)
	{
		pimgline[diblinenum] = (png_bytep)(pdibdata + (dibheight - diblinenum - 1) * dibPitch);
	}
	
	png_write_image(ppng, pimgline);
	free ((char *) pimgline);
	png_write_end(ppng, pipng);
	png_destroy_write_struct(&ppng, &pipng);
	return TRUE;
}
