#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//BITMAPFILEHEADER  ptr index
#define BF_TYPE 0
#define BF_SIZE 2
#define BF_RESERVED1 6
#define BF_RESERVED2 8
#define BF_OFFBITS 10

//WINBMPINFOHEADER  ptr index
#define BI_SIZE 14
#define BI_WIDTH 18
#define BI_HEIGHT 22
#define BI_PLANES 26
#define BI_BITCOUNT 28
#define BI_COMPRESSION 30
#define BI_SIZEIMAGE 34
#define BI_XPELSPERMETER 38
#define BI_YPELSPERMETER 42
#define BI_CLRUSED 46
#define BI_CLRIMPORTANT 50

//Palette entry  ptr index
#define BF_PALETTE 54

//DIB  ptr index
#define BF_DIBP8 1078
#define BF_DIBP24 54

#undef TRUE
#undef FALSE
#define TRUE 1
#define FALSE 0

static unsigned char CheckBitmapfileHeader(unsigned char *pBitmapImage)
{
	//Check bitmap file header.
	if(!pBitmapImage ||
		*(unsigned short *)(&pBitmapImage[BF_TYPE]) != 0x4D42 ||
		*(unsigned short *)(&pBitmapImage[BF_RESERVED1]) != 0 ||
		*(unsigned short *)(&pBitmapImage[BF_RESERVED2]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_SIZE]) != 40 ||
		*(unsigned short *)(&pBitmapImage[BI_PLANES]) != 1 ||
		*(unsigned long *)(&pBitmapImage[BI_COMPRESSION]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_XPELSPERMETER]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_YPELSPERMETER]) != 0 ||
		!((*(unsigned long *)(&pBitmapImage[BF_OFFBITS]) == 54 && *(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 24 &&
		*(unsigned long *)(&pBitmapImage[BI_CLRUSED]) == 0 && *(unsigned long *)(&pBitmapImage[BI_CLRIMPORTANT]) == 0) ||
		(*(unsigned long *)(&pBitmapImage[BF_OFFBITS]) == 1078 && *(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 8 &&
		*(unsigned long *)(&pBitmapImage[BI_CLRUSED]) == 256 && *(unsigned long *)(&pBitmapImage[BI_CLRIMPORTANT]) == 256)))
		return FALSE;
	return TRUE;
}

unsigned char *ConvertBitmapImage24To8IfWithin(unsigned char *pBitmapImage)
{
	//Check 24 bit bitmap file header.
	if(!pBitmapImage ||
		*(unsigned short *)(&pBitmapImage[BF_TYPE]) != 0x4D42 ||
		*(unsigned short *)(&pBitmapImage[BF_RESERVED1]) != 0 ||
		*(unsigned short *)(&pBitmapImage[BF_RESERVED2]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_SIZE]) != 40 ||
		*(unsigned short *)(&pBitmapImage[BI_PLANES]) != 1 ||
		*(unsigned long *)(&pBitmapImage[BI_COMPRESSION]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_XPELSPERMETER]) != 0 ||
		*(unsigned long *)(&pBitmapImage[BI_YPELSPERMETER]) != 0 ||
		!(*(unsigned long *)(&pBitmapImage[BF_OFFBITS]) == 54 && *(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 24 &&
		*(unsigned long *)(&pBitmapImage[BI_CLRUSED]) == 0 && *(unsigned long *)(&pBitmapImage[BI_CLRIMPORTANT]) == 0))
		return pBitmapImage;
	
	unsigned long bmfWidth = *(unsigned long *)(&pBitmapImage[BI_WIDTH]);
	unsigned long bmfHeight = *(unsigned long *)(&pBitmapImage[BI_HEIGHT]);
	unsigned char *pbmp8 = (unsigned char *) malloc (14 + 40 + 1024 + bmfWidth * bmfHeight);
	
	if(!pbmp8)
		return pBitmapImage;

	//BITMAPFILEHEADER (14 byte)
	*(unsigned short *)(&pbmp8[BF_TYPE]) = 0x4D42;
	*(unsigned long *)(&pbmp8[BF_SIZE]) = 54 + 1024 + bmfWidth * bmfHeight;
	*(unsigned short *)(&pbmp8[BF_RESERVED1]) = 0;
	*(unsigned short *)(&pbmp8[BF_RESERVED2]) = 0;
	*(unsigned long *)(&pbmp8[BF_OFFBITS]) = 1078;

	//WINBMPINFOHEADER (40 byte)
	*(unsigned long *)(&pbmp8[BI_SIZE]) = 40;
	*(unsigned long *)(&pbmp8[BI_WIDTH]) = bmfWidth;
	*(unsigned long *)(&pbmp8[BI_HEIGHT]) = bmfHeight;
	*(unsigned short *)(&pbmp8[BI_PLANES]) = 1;
	*(unsigned short *)(&pbmp8[BI_BITCOUNT]) = 8;
	*(unsigned long *)(&pbmp8[BI_COMPRESSION]) = 0;
	*(unsigned long *)(&pbmp8[BI_SIZEIMAGE]) = bmfWidth * bmfHeight;
	*(unsigned long *)(&pbmp8[BI_XPELSPERMETER]) = 0;
	*(unsigned long *)(&pbmp8[BI_YPELSPERMETER]) = 0;
	*(unsigned long *)(&pbmp8[BI_CLRUSED]) = 256;
	*(unsigned long *)(&pbmp8[BI_CLRIMPORTANT]) = 256;

	memset((void *)&pbmp8[BF_PALETTE], 0, 1024 + *(unsigned long *)(&pbmp8[BI_SIZEIMAGE]));
	unsigned long i;
	unsigned long dbxy = *(unsigned long *)(&pbmp8[BI_SIZEIMAGE]);
	unsigned long pali = 0;
	unsigned long dbi = 0;
	unsigned char tmppal[4];
	tmppal[3] = 0;
	for(; pali < 256 && dbi < dbxy; dbi++) {
		tmppal [0] = pBitmapImage[BF_DIBP24 + dbi * 3];
		tmppal [1] = pBitmapImage[BF_DIBP24 + dbi * 3 + 1];
		tmppal [2] = pBitmapImage[BF_DIBP24 + dbi * 3 + 2];
		i = 0;
		do {
			if(i == pali) {
				*(unsigned long *)(&pbmp8[BF_PALETTE + pali * 4]) = *(unsigned long *)(&tmppal[0]);
				pali++;
				break;
			}
			else if(i < pali && *(unsigned long *)(&pbmp8[BF_PALETTE + i * 4]) == *(unsigned long *)(&tmppal[0]))
				break;
		} while(i++ < pali);
	}

	if(pali >= 256 && dbi < dbxy) {
		free ((char *) pbmp8);
		return pBitmapImage;
	}
	unsigned char tmppix[4];
	tmppix[3] = 0;
	dbi = 0;
	for(; dbi < dbxy; dbi++) {
		tmppix [0] = pBitmapImage[BF_DIBP24 + dbi * 3];
		tmppix [1] = pBitmapImage[BF_DIBP24 + dbi * 3 + 1];
		tmppix [2] = pBitmapImage[BF_DIBP24 + dbi * 3 + 2];
		i = 0;
		do {
			if(*(unsigned long *)(&pbmp8[BF_PALETTE + i * 4]) == *(unsigned long *)(&tmppix[0])) {
				pbmp8[BF_DIBP8 + dbi] = (unsigned char)i;
				break;
			}
			else if(i == pali) {
				break;
			}
		} while(i++ < pali);
	}
	free ((char *) pBitmapImage);
	return pbmp8;
}

// not supported yet.
unsigned char *ConvertBitmapImage24To8(unsigned char *pBitmapImage)
{
	return pBitmapImage;
}

unsigned char *x2wBitmapImage(unsigned char *pBitmapImage)
{
	//Check bitmap file header.
	if(!CheckBitmapfileHeader(pBitmapImage))
		return pBitmapImage;
	unsigned char *px2wbmp = (unsigned char *) malloc ((*(unsigned long *)(&pBitmapImage[BF_SIZE])) + (*(unsigned long *)(&pBitmapImage[BI_SIZEIMAGE])));
	if(!px2wbmp)
		return pBitmapImage;
	memmove(px2wbmp, pBitmapImage, *(unsigned long *)(&pBitmapImage[BF_OFFBITS]));
	
	*(unsigned long *)(&px2wbmp[BF_SIZE]) = (*(unsigned long *)(&pBitmapImage[BF_SIZE])) + (*(unsigned long *)(&pBitmapImage[BI_SIZEIMAGE]));
	*(unsigned long *)(&px2wbmp[BI_SIZEIMAGE]) = (*(unsigned long *)(&pBitmapImage[BI_SIZEIMAGE])) << 1;
	*(unsigned long *)(&px2wbmp[BI_WIDTH]) = (*(unsigned long *)(&pBitmapImage[BI_WIDTH])) << 1;
	
	unsigned long srci = *(unsigned long *)(&pBitmapImage[BF_OFFBITS]);
	unsigned long dsti = *(unsigned long *)(&pBitmapImage[BF_OFFBITS]);
	unsigned long srcbmpfsize = *(unsigned long *)(&pBitmapImage[BF_SIZE]);

	if(*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) == 24) {
		for(; srci < srcbmpfsize; srci += 3, dsti += 6) {
			px2wbmp[dsti] = pBitmapImage[srci];
			px2wbmp[dsti + 1] = pBitmapImage[srci + 1];
			px2wbmp[dsti + 2] = pBitmapImage[srci + 2];
			px2wbmp[dsti + 3] = pBitmapImage[srci];
			px2wbmp[dsti + 4] = pBitmapImage[srci + 1];
			px2wbmp[dsti + 5] = pBitmapImage[srci + 2];
		}
	}
	else {
		for(; srci < srcbmpfsize; srci += 1, dsti += 2) {
			px2wbmp[dsti] = pBitmapImage[srci];
			px2wbmp[dsti + 1] = pBitmapImage[srci];
		}
	}
	free ((char *) pBitmapImage);
	return px2wbmp;
}

unsigned char *x2hBitmapImage(unsigned char *pBitmapImage)
{
	//Check bitmap file header.
	if(!CheckBitmapfileHeader(pBitmapImage))
		return pBitmapImage;
	unsigned char *px2hbmp = (unsigned char *) malloc ((*(unsigned long *)(&pBitmapImage[BF_SIZE])) + (*(unsigned long *)(&pBitmapImage[BI_SIZEIMAGE])));
	if(!px2hbmp)
		return pBitmapImage;
	memmove(px2hbmp, pBitmapImage, *(unsigned long *)(&pBitmapImage[BF_OFFBITS]));
	
	*(unsigned long *)(&px2hbmp[BF_SIZE]) = (*(unsigned long *)(&pBitmapImage[BF_SIZE])) + (*(unsigned long *)(&pBitmapImage[BI_SIZEIMAGE]));
	*(unsigned long *)(&px2hbmp[BI_SIZEIMAGE]) = (*(unsigned long *)(&pBitmapImage[BI_SIZEIMAGE])) << 1;
	*(unsigned long *)(&px2hbmp[BI_HEIGHT]) = (*(unsigned long *)(&pBitmapImage[BI_HEIGHT])) << 1;
	
	unsigned long srci = *(unsigned long *)(&pBitmapImage[BF_OFFBITS]);
	unsigned long dsti = *(unsigned long *)(&pBitmapImage[BF_OFFBITS]);
	unsigned long srcbmpfsize = *(unsigned long *)(&pBitmapImage[BF_SIZE]);
	unsigned long srcdibpitch = (*(unsigned long *)(&pBitmapImage[BI_SIZEIMAGE])) / (*(unsigned long *)(&pBitmapImage[BI_HEIGHT]));
	unsigned long srcdibpitchx2 = srcdibpitch << 1;
	for(; srci < srcbmpfsize; srci += srcdibpitch, dsti += srcdibpitchx2) {
		memmove(&px2hbmp[dsti], &pBitmapImage[srci], srcdibpitch);
		memmove(&px2hbmp[dsti + srcdibpitch], &pBitmapImage[srci], srcdibpitch);
	}
	free ((char *) pBitmapImage);
	return px2hbmp;
}


/**********************
  RLE compression code
***********************/

#define BI_RLE8       1
#define BI_RLE4       2

static unsigned int EncodeImage(unsigned int imagerows, const unsigned int bytes_per_line,
  const unsigned char *pixels,unsigned char *compressed_pixels)
{
  int
    y;

  register const unsigned char
    *p;

  register int
    i,
    x;

  register unsigned char
    *q;

  /*
    Runlength encode pixels.
  */
  p=pixels;
  q=compressed_pixels;
  i=0;
  for (y=0; y < (int) imagerows; y++)
  {
    for (x=0; x < (int) bytes_per_line; x+=i)
    {
      /*
        Determine runlength.
      */
      for (i=1; ((x+i) < (int) bytes_per_line); i++)
        if ((*(p+i) != *p) || (i == 255))
          break;
      *q++=(unsigned char) i;
      *q++=(*p);
      p+=i;
    }
    /*
      End of line.
    */
    *q++=0x00;
    *q++=0x00;
  }
  /*
    End of bitmap.
  */
  *q++=0;
  *q++=0x01;
  return((unsigned int) (q-compressed_pixels));
}

unsigned char *bmp_compress_dib_rle(unsigned char *pBitmapImage)
{
	int row_bytes, num_palette;
	unsigned int data_size;
	unsigned char *new_dib;
	unsigned char *new_bmpf;
	unsigned char *src, *dest;

	if(*(unsigned long *)(&pBitmapImage[BI_COMPRESSION]) != 0)
		return NULL;
	if((*(unsigned short *)(&pBitmapImage[BI_BITCOUNT]) != 8))
		return NULL;
	row_bytes = ((*(unsigned long *)(&pBitmapImage[BI_WIDTH])) * (*(unsigned short *)(&pBitmapImage[BI_BITCOUNT])) / 8);
	num_palette = (int)(*(unsigned long *)(&pBitmapImage[BI_CLRUSED]));
	if(num_palette == 0 )
		num_palette = (1 << (*(unsigned short *)(&pBitmapImage[BI_BITCOUNT])));

	new_bmpf = (unsigned char *) malloc (14 + 40 + (num_palette * 4) + ((*(unsigned long *)(&pBitmapImage[BI_HEIGHT]) + 2) * (row_bytes + 2) * 2) + 2);
	
	if(!new_bmpf)
		return NULL;
	new_dib = &new_bmpf[14];
	memcpy(new_dib, &pBitmapImage[14], (40 + num_palette * 4));

	src  = (&pBitmapImage[14] + 40 + num_palette * 4);
	dest = (new_dib +  40 + num_palette * 4);

	*(unsigned long *)(&new_dib[BI_COMPRESSION - 14]) = BI_RLE8;

	data_size = EncodeImage((unsigned int)(*(unsigned long *)(&pBitmapImage[BI_HEIGHT])), (const unsigned int)row_bytes, (const unsigned char *)src, (unsigned char *)dest);

	*(unsigned long *)(&new_dib[BI_CLRUSED - 14]) = (unsigned long)num_palette;
	*(unsigned long *)(&new_dib[BI_SIZEIMAGE - 14]) = (unsigned long)data_size;
	memcpy(new_bmpf, pBitmapImage, 14);
	*(unsigned long *)(&new_bmpf[BF_SIZE]) = (54 + 1024 + data_size);
	return new_bmpf;
}
