#include <stdio.h>
#include "../snes9x/snes9x.h"
#include "../snes9x/apu.h"
#include "../snes9x/spc700.h"
#include "../snes9x/soundux.h"

extern "C" void S9xReset();

extern int NoiseFreq [32];
extern void S9xSetEnvRate (Channel *ch, unsigned long rate, int direction, int target);

SAPURegisters BackupAPURegisters;
unsigned char BackupAPURAM[65536];
unsigned char BackupAPUExtraRAM[64];
unsigned char BackupDSPRAM[128];

#define C32PM (2048000/50)/32
#define ABS(a) ((a) < 0 ? -(a) : (a))
#define ENVX_SHIFT 24


// The callback
void DoTimer(){
	if (APU.TimerEnabled [2])
	{
		APU.Timer [2]++;
		if (APU.Timer [2] >= APU.TimerTarget [2])
		{
			IAPU.RAM [0xff] = (IAPU.RAM [0xff] + 1) & 0xf;
			APU.Timer [2] = 0;
#ifdef SPC700_SHUTDOWN          
			IAPU.WaitCounter++;
			IAPU.APUExecuting = TRUE;
#endif          
		}
	}
	if (IAPU.TimerErrorCounter>=8)
	{
		IAPU.TimerErrorCounter=0;
		if (APU.TimerEnabled [0])
		{
			APU.Timer [0]++;
			if (APU.Timer [0] >= APU.TimerTarget [0])
			{
				IAPU.RAM [0xfd] = (IAPU.RAM [0xfd] + 1) & 0xf;
				APU.Timer [0] = 0;
#ifdef SPC700_SHUTDOWN          
				IAPU.WaitCounter++;
				IAPU.APUExecuting = TRUE;
#endif          
			}
		}
		if (APU.TimerEnabled [1])
		{
			APU.Timer [1]++;
			if (APU.Timer [1] >= APU.TimerTarget [1])
			{
				IAPU.RAM [0xfe] = (IAPU.RAM [0xfe] + 1) & 0xf;
				APU.Timer [1] = 0;
#ifdef SPC700_SHUTDOWN          
				IAPU.WaitCounter++;
				IAPU.APUExecuting = TRUE;
#endif          
			}
		}
	}
}

extern "C" void S9xSyncSpeed( void);

void SPCPlayMainLoop(void)
{
	// APU_LOOP
	for(int c=0;c<C32PM;c++){
		for(int ic=0;ic<32;ic++){
		APU_EXECUTE1();
		}
		IAPU.TimerErrorCounter++;
		DoTimer();
	}
	S9xSyncSpeed();
}

/*
#ifdef __WIN32__
DWORD WINAPI SPCPlayMainLoopThread(LPVOID data)
{
	PDATA pData;
	pData = (PDATA)data;
	while (!pData->thread_end) {
	// APU_LOOP
		for(int c=0;c<C32PM;c++){
			for(int ic=0;ic<32;ic++){
			APU_EXECUTE1();
			}
			IAPU.TimerErrorCounter++;
			DoTimer();
		}
		S9xSyncSpeed();
	}
	return 0;
}
#endif
*/

EXTERN_C uint8 APUROM [64];

void SPCSetAPUControl (uint8 byte)
{
    if ((byte & 1) != 0 && !APU.TimerEnabled [0])
    {
	APU.Timer [0] = 0;
	IAPU.RAM [0xfd] = 0;
	if ((APU.TimerTarget [0] = IAPU.RAM [0xfa]) == 0)
	    APU.TimerTarget [0] = 0x100;
    }
    if ((byte & 2) != 0 && !APU.TimerEnabled [1])
    {
	APU.Timer [1] = 0;
	IAPU.RAM [0xfe] = 0;
	if ((APU.TimerTarget [1] = IAPU.RAM [0xfb]) == 0)
	    APU.TimerTarget [1] = 0x100;
    }
    if ((byte & 4) != 0 && !APU.TimerEnabled [2])
    {
	APU.Timer [2] = 0;
	IAPU.RAM [0xff] = 0;
	if ((APU.TimerTarget [2] = IAPU.RAM [0xfc]) == 0)
	    APU.TimerTarget [2] = 0x100;
    }
    APU.TimerEnabled [0] = byte & 1;
    APU.TimerEnabled [1] = (byte & 2) >> 1;
    APU.TimerEnabled [2] = (byte & 4) >> 2;

#if 0
//by kinoshita
    if (byte & 0x10)
	IAPU.RAM [0xF4] = IAPU.RAM [0xF5] = 0;

    if (byte & 0x20)
	IAPU.RAM [0xF6] = IAPU.RAM [0xF7] = 0;
#endif

    if (byte & 0x80)
    {
	if (!APU.ShowROM)
	{
	    memmove (&IAPU.RAM [0xffc0], APUROM, sizeof (APUROM));
	    APU.ShowROM = TRUE;
	}
    }
    else
    {
	if (APU.ShowROM)
	{
	    APU.ShowROM = FALSE;
	    memmove (&IAPU.RAM [0xffc0], APU.ExtraRAM, sizeof (APUROM));
	}
    }
    IAPU.RAM [0xf1] = byte;
}

// Restore SPC Load State
void RestoreSPC(){
	int i;

	APURegisters.PC = BackupAPURegisters.PC;
	APURegisters.YA.B.A = BackupAPURegisters.YA.B.A;
	APURegisters.X = BackupAPURegisters.X;
	APURegisters.YA.B.Y = BackupAPURegisters.YA.B.Y;
	APURegisters.P = BackupAPURegisters.P;
	APURegisters.S = BackupAPURegisters.S;
	memcpy(IAPU.RAM,BackupAPURAM,65536);
	memcpy(APU.ExtraRAM,BackupAPUExtraRAM,64);
	memcpy(APU.DSP,BackupDSPRAM,128);

	for(i=0;i<4;i++){
		APU.OutPorts [i] = IAPU.RAM[0xf4+i];
	}
	IAPU.TimerErrorCounter=0;
	for (i=0;i<3;i++){
		if (IAPU.RAM[0xfa+i] == 0)
			APU.TimerTarget[i] = 0x100;
		else
			APU.TimerTarget[i] = IAPU.RAM[0xfa+i];
	}
	SPCSetAPUControl(IAPU.RAM[0xf1]);
	IAPU.PC = IAPU.RAM + APURegisters.PC;
	S9xAPUUnpackStatus();
	if (APUCheckDirectPage ())
		IAPU.DirectPage = IAPU.RAM + 0x100;
	else
		IAPU.DirectPage = IAPU.RAM;
	IAPU.APUExecuting = TRUE;

	S9xSetFrequencyModulationEnable(APU.DSP[APU_PMON]);
	S9xSetMasterVolume((signed char)APU.DSP[APU_MVOL_LEFT], (signed char)APU.DSP[APU_MVOL_RIGHT]);
	S9xSetEchoVolume((signed char)APU.DSP[APU_EVOL_LEFT], (signed char)APU.DSP[APU_EVOL_RIGHT]);

	uint8 mask = 1;
	for(i=0;i<8;i++,mask<<=1){
		if (APU.DSP [APU_NON] & mask){
			S9xSetSoundType (i, SOUND_NOISE);
		}else{
			S9xSetSoundType (i, SOUND_SAMPLE);
		}
		S9xSetSoundHertz(i,(int)(((((int)APU.DSP[APU_P_HIGH+(i<<4)]<<8) + APU.DSP[APU_P_LOW+(i<<4)])&FREQUENCY_MASK)*8));
		S9xFixEnvelope (i, APU.DSP[APU_GAIN+(i<<4)], APU.DSP[APU_ADSR1+(i<<4)], APU.DSP[APU_ADSR2+(i<<4)]);
		SoundData.channels[i].sample_number = (uint16)APU.DSP[APU_SRCN+(i<<4)];
		if (APU.DSP[APU_KON] & mask)
		{
			APU.KeyedChannels |= mask;
			S9xPlaySample (i);
		}
		else
			S9xSetSoundKeyOff (i);
		S9xSetSoundVolume (i, (signed char)APU.DSP[APU_VOL_LEFT+(i<<4)], (signed char)APU.DSP[APU_VOL_RIGHT+(i<<4)]);
	}
	S9xSetEchoEnable (APU.DSP [APU_EON]);
	S9xFixSoundAfterSnapshotLoad();
}

// Load SPC file
bool LoadSPC(char *pszFileName)
{
	FILE *file;
	char temp[64];
	if((file=fopen(pszFileName,"rb"))!=NULL)
	{
		Options.APURAMInitialValue = 0xff;
		Options.APURAMInitialValue0 = 0xff;
		Options.SRAMInitialValue = 0xaa;
		Options.SRAMInitialValue0 = 0xaa;
		memset(&SNESGameFixes, 0, sizeof(SNESGameFixes));
//		S9xResetAPU();
		S9xReset();
		Settings.FrameTime = 20;
		fseek(file,37,SEEK_SET);
		fread(&BackupAPURegisters.PC,2,1,file);
		fread(&BackupAPURegisters.YA.B.A,1,1,file);
		fread(&BackupAPURegisters.X,1,1,file);
		fread(&BackupAPURegisters.YA.B.Y,1,1,file);
		fread(&BackupAPURegisters.P,1,1,file);
		fread(&BackupAPURegisters.S,1,1,file);
		fseek(file,256,SEEK_SET);
		fread(BackupAPURAM,65536,1,file);
		fread(BackupDSPRAM,128,1,file);
		fread(temp,64,1,file);
		fread(BackupAPUExtraRAM,64,1,file);

		fclose(file);
		RestoreSPC();
		return 0;
	}
	return 1;
}
