#include "../snes9x/memmap.h"
#include <commdlg.h>
#include <shlobj.h>
#include "cfgdlg.h"

extern char ROMFileName[];
extern char InitialROMFileDirectory[];
extern char ScreenShotDirectory[];
extern char SPCDumpDirectory[];
extern void SlashAdd(char *PathStr);
extern void SlashRemove(char *PathStr);
extern void RemoveFileName(char *PathStr);

extern "C" const char *S9xBasename (const char *f);
LRESULT CALLBACK BrowseHook(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam);
BOOL BIOSOpenFile(HWND hWnd, char *FileName, char *FileTitle, char *InitialROMFileDirectory);

static char InitialROMFileDirectory_cpy[_MAX_PATH];
static char GUI_FreezeFileDir_cpy[_MAX_PATH];
static char ScreenShotDirectory_cpy[_MAX_PATH];
static char SPCDumpDirectory_cpy[_MAX_PATH];
static char Memory_BSXBIOSImageFileName_cpy[_MAX_PATH];
static char InitialBIOSFileDirectory[_MAX_PATH];
static char BIOSFileTitle[_MAX_PATH];

void DirectorySettingsLoad()
{
	lstrcpy(InitialROMFileDirectory_cpy, InitialROMFileDirectory);
	lstrcpy(GUI_FreezeFileDir_cpy, GUI.FreezeFileDir);
	lstrcpy(ScreenShotDirectory_cpy, ScreenShotDirectory);
	lstrcpy(SPCDumpDirectory_cpy, SPCDumpDirectory);
	lstrcpy(Memory_BSXBIOSImageFileName_cpy, Memory.BSXBIOSImageFileName);
	lstrcpy(InitialBIOSFileDirectory, Memory_BSXBIOSImageFileName_cpy);
	RemoveFileName(InitialBIOSFileDirectory);
	lstrcpy(BIOSFileTitle, S9xBasename((const char *)Memory_BSXBIOSImageFileName_cpy));
}

void DirectorySettingsSave()
{
	if(lstrcmp(InitialROMFileDirectory, InitialROMFileDirectory_cpy) && lstrlen(InitialROMFileDirectory_cpy) <= (_MAX_PATH - 7)) {
		lstrcpy(InitialROMFileDirectory, InitialROMFileDirectory_cpy);
		lstrcpy(ROMFileName, InitialROMFileDirectory_cpy);
		SlashRemove(ROMFileName);
		lstrcat(ROMFileName, "\\a.smc");
	}
	lstrcpy(GUI.FreezeFileDir, GUI_FreezeFileDir_cpy);
	lstrcpy(ScreenShotDirectory, ScreenShotDirectory_cpy);
	lstrcpy(SPCDumpDirectory, SPCDumpDirectory_cpy);
	lstrcpy(Memory.BSXBIOSImageFileName, Memory_BSXBIOSImageFileName_cpy);
}

LRESULT CALLBACK DirectoryDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
	static char tmpstr[_MAX_PATH];
	switch (uMsg) {
		case WM_INITDIALOG:
			lstrcpy(tmpstr, InitialROMFileDirectory_cpy);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_ROM), tmpstr);
			lstrcpy(tmpstr, GUI_FreezeFileDir_cpy);
			SlashAdd(tmpstr);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_SAVE), tmpstr);
			lstrcpy(tmpstr, ScreenShotDirectory_cpy);
			SlashAdd(tmpstr);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_SCREEN), tmpstr);
			lstrcpy(tmpstr, SPCDumpDirectory_cpy);
			SlashAdd(tmpstr);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_SPC), tmpstr);
			SetWindowText(GetDlgItem(hWnd, IDC_DIR_BSXBIOS), Memory_BSXBIOSImageFileName_cpy);
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_DIR_ROM:
				case IDC_DIR_SAVE:
				case IDC_DIR_SCREEN:
				case IDC_DIR_SPC:
				case IDC_DIR_BSXBIOS:
					if(HIWORD(wParam) == EN_CHANGE) {
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_BUTTON_DIR_ROM:
				case IDC_BUTTON_DIR_SAVE:
				case IDC_BUTTON_DIR_SCREEN:
				case IDC_BUTTON_DIR_SPC:
					{
						char BrowseDir[_MAX_PATH];
						char InitialBrowseDir[_MAX_PATH];
						int idctrl = IDC_DIR_ROM;
						BROWSEINFO bi;
						FillMemory(&bi, sizeof(BROWSEINFO), 0x00);

						switch(LOWORD(wParam))
						{
							case IDC_BUTTON_DIR_ROM:
								bi.lpszTitle = (GUI.Language ? "ǂݍ݃tH_I" : "Select ROM Folder");
								idctrl = IDC_DIR_ROM;
								lstrcpy(InitialBrowseDir, InitialROMFileDirectory_cpy);
								break;

							case IDC_BUTTON_DIR_SAVE:
								bi.lpszTitle = (GUI.Language ? "Z[ut@CۑtH_I" : "Select Save Folder");
								idctrl = IDC_DIR_SAVE;
								lstrcpy(InitialBrowseDir, GUI_FreezeFileDir_cpy);
								break;

							case IDC_BUTTON_DIR_SCREEN:
								bi.lpszTitle = (GUI.Language ? "XN[VbgۑtH_I" : "Select Screen shot Save Folder");
								idctrl = IDC_DIR_SCREEN;
								lstrcpy(InitialBrowseDir, ScreenShotDirectory_cpy);
								break;

							case IDC_BUTTON_DIR_SPC:
								bi.lpszTitle = (GUI.Language ? "SPCۑtH_I" : "Select SPC Save Folder");
								idctrl = IDC_DIR_SPC;
								lstrcpy(InitialBrowseDir, SPCDumpDirectory_cpy);
								break;
						}
						SlashAdd(InitialBrowseDir);
						bi.hwndOwner = hWnd;
						bi.ulFlags = BIF_RETURNONLYFSDIRS | BIF_DONTGOBELOWDOMAIN;
						LPITEMIDLIST lpiil;
						bi.lpfn = (BFFCALLBACK)BrowseHook;
						bi.lParam=(LPARAM)InitialBrowseDir;
						
						if((lpiil = SHBrowseForFolder(&bi)))
						{
							if( SHGetPathFromIDList( lpiil, BrowseDir) ) {
								SetWindowText(GetDlgItem(hWnd, idctrl), BrowseDir);
								//ROMǂݍݏfBNgȊOŌ̃pX؂폜
								if(idctrl != IDC_DIR_ROM)
									SlashRemove(BrowseDir);
								if(idctrl == IDC_DIR_ROM)
									lstrcpy(InitialROMFileDirectory_cpy, BrowseDir);
								else if(idctrl == IDC_DIR_SAVE)
									lstrcpy(GUI_FreezeFileDir_cpy, BrowseDir);
								else if(idctrl == IDC_DIR_SCREEN)
									lstrcpy(ScreenShotDirectory_cpy, BrowseDir);
								else if(idctrl == IDC_DIR_SPC)
									lstrcpy(SPCDumpDirectory_cpy, BrowseDir);
							}
							CoTaskMemFree(lpiil );
						}
						break;
					}
					break;

				case IDC_BUTTON_DIR_BSXBIOS:
					if (BIOSOpenFile(hWnd, Memory_BSXBIOSImageFileName_cpy, BIOSFileTitle, InitialBIOSFileDirectory)) {
						SetWindowText(GetDlgItem(hWnd, IDC_DIR_BSXBIOS), Memory_BSXBIOSImageFileName_cpy);
						lstrcpy(InitialBIOSFileDirectory, Memory_BSXBIOSImageFileName_cpy);
						RemoveFileName(InitialBIOSFileDirectory);
						lstrcpy(BIOSFileTitle, S9xBasename((const char *)Memory_BSXBIOSImageFileName_cpy));
					}
					break;

				default:
					break;
			}
			break;

		case WM_CTLCOLORSTATIC:
			if((HWND)lParam == GetDlgItem(hWnd, IDC_DIR_ROM) || (HWND)lParam == GetDlgItem(hWnd, IDC_DIR_SAVE)
				|| (HWND)lParam == GetDlgItem(hWnd, IDC_DIR_SCREEN) || (HWND)lParam == GetDlgItem(hWnd, IDC_DIR_SPC)
				|| (HWND)lParam == GetDlgItem(hWnd, IDC_DIR_BSXBIOS))
				return (long)GetStockObject(WHITE_BRUSH);
			break;

		case WM_NOTIFY:
			switch(((NMHDR *)lParam)->code) {
				case PSN_APPLY:
					ConfigDlgSettingsSave();
					break;

				case PSN_HELP:
					ConfigDlgDefaultSettingsLoad(hWnd);
					break;

				case PSN_QUERYCANCEL:
					break;

				case PSN_SETACTIVE:
					StartPageIndex = 3;
					break;

				default:
					break;
			}
			break;

		default:
			break;
	}
	return FALSE;
}

BOOL BIOSOpenFile(HWND hWnd, char *FileName, char *FileTitle, char *InitialROMFileDirectory)
{
	OPENFILENAME ofn;
	FillMemory((PVOID)&ofn, sizeof(ofn), 0);
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = hWnd;
	ofn.lpstrFilter = 
	"Snes(*.smc, *.sfc, *.swc, *.fig, *.078)\0*.smc;*.sfc;*.swc;*.fig;*.078\0"
	"All files(*.*)\0*.*\0\0";
	ofn.lpstrFile = FileName;
	ofn.nMaxFile = _MAX_PATH;
	ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_NONETWORKBUTTON | OFN_EXPLORER;//OFN_ENABLEHOOK
	ofn.lpstrDefExt = "smc";
	ofn.nMaxFileTitle = _MAX_PATH;
	ofn.lpstrFileTitle = FileTitle;
	ofn.lpstrTitle = (GUI.Language ? "BS-X BIOS t@CI" : "Select BS-X BIOS FileName");
//	ofn.lpfnHook = (LPOFNHOOKPROC)GetOpenFileNameHook;
	ofn.lpstrInitialDir = InitialROMFileDirectory;
	return GetOpenFileName(&ofn);
}

LRESULT CALLBACK BrowseHook(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
//	static DWORD Style;
	switch (uMsg) {
		case BFFM_INITIALIZED:
			SendMessage(hWnd,BFFM_SETSELECTION,(WPARAM)TRUE,lParam);
#if 0
			Style = GetWindowLong(hWnd, GWL_EXSTYLE );
			Style &= ~(WS_EX_CONTEXTHELP);
			SetWindowLong(hWnd, GWL_EXSTYLE, Style);
#endif
			break;

		default:
			break;
	}
	return FALSE;
}
